from .topology import Topology
from .elements import Atom, Bond, Angle, Dihedral, Proper, Improper
from .molecule import Molecule
from .zmatrix import ZMatrix
from .rotamer import RotamerLibrary
from .conformer import BCEConformations
from .mapper import Mapper
from .alchemistry import Alchemizer



"""
This module contains classes and methods related with alchemical modifications for molecular topologies.
"""


from abc import ABC


class Alchemizer(object):
    """
    It defines the Alchemizer class.
    """

    def __init__(self, topology1, topology2):
        """
        It initializes an Alchemizer object, which generates alchemical
        representations considering the topologies of two different
        molecules.

        Parameters
        ----------
        topology1 : a peleffy.topology.Topology object
            The molecular topology representation of molecule 1
        topology2 : a peleffy.topology.Topology object
            The molecular topology representation of molecule 2
        """

        # Check topologies
        import peleffy

        if (not isinstance(topology1, peleffy.topology.Topology)
                and not
                isinstance(topology1, peleffy.topology.topology.Topology)):
            raise TypeError('Invalid input topology 1')

        if (not isinstance(topology2, peleffy.topology.Topology)
                and not
                isinstance(topology2, peleffy.topology.topology.Topology)):
            raise TypeError('Invalid input topology 2')

        self._topology1 = topology1
        self._topology2 = topology2
        self._molecule1 = topology1.molecule
        self._molecule2 = topology2.molecule

        from peleffy.topology import Mapper

        # Map atoms from both molecules
        self._mapper = Mapper(self.molecule1, self.molecule2,
                              include_hydrogens=True)

        self._mapping = self._mapper.get_mapping()
        self._mcs_mol = self._mapper.get_mcs()

        # Join the two topologies
        self._joint_topology, self._non_native_atoms, \
            self._non_native_bonds, self._non_native_angles, \
            self._non_native_propers, self._non_native_impropers,\
            self._mol2_to_alc_map = self._join_topologies()

        # Get exclusive topological elements in topology 1
        self._exclusive_atoms, self._exclusive_bonds, \
        self._exclusive_angles, self._exclusive_propers, \
        self._exclusive_impropers = self._get_exclusive_elements()

        # Find connections
        self._connections = self._find_connections()

        # Generate alchemical graph
        self._graph, self._rotamers = self._generate_alchemical_graph()

        # Assign PDB atom names
        self._assign_pdb_atom_names()

    @property
    def topology1(self):
        """
        It returns the first peleffy's Topology.

        Returns
        -------
        topology1 : a peleffy.topology.Topology
            The first peleffy's Topology object
        """
        return self._topology1

    @property
    def topology2(self):
        """
        It returns the second peleffy's Topology.

        Returns
        -------
        topology2 : a peleffy.topology.Topology
            The second peleffy's Topology object
        """
        return self._topology2

    @property
    def molecule1(self):
        """
        It returns the first molecule.

        Returns
        -------
        molecule1 : a peleffy.topology.Molecule
            The first Molecule object
        """
        return self._molecule1

    @property
    def molecule2(self):
        """
        It returns the second molecule.

        Returns
        -------
        molecule2 : a peleffy.topology.Molecule
            The second Molecule object
        """
        return self._molecule2

    @property
    def mapping(self):
        """
        It returns the mapping between both molecules.

        Returns
        -------
        mapping : list[tuple]
            The list of atom pairs between both molecules, represented
            with tuples
        """
        return self._mapping

    @property
    def mcs_mol(self):
        """
        It returns the Maximum Common Substructure (MCS) between
        both molecules.

        Parameters
        ----------
        mcs_mol : an RDKit.molecule object
            The resulting MCS molecule
        """
        return self._mcs_mol

    @property
    def connections(self):
        """
        It returns the list of connections between molecule 1 and non
        native atoms of molecule 2.

        Returns
        -------
        connections : list[tuple[int, int]]
            The list of connections between molecule 1 and non
            native atoms of molecule 2
        """
        return self._connections

    def get_alchemical_topology(self, fep_lambda=None, coul_lambda=None,
                                coul1_lambda=None, coul2_lambda=None,
                                vdw_lambda=None, bonded_lambda=None):
        """
        Given a lambda, it returns an alchemical topology after
        combining both input topologies.

        Parameters
        ----------
        fep_lambda : float
            The value to define an FEP lambda. This lambda affects
            all the parameters. It needs to be contained between
            0 and 1. Default is None
        coul_lambda : float
            The value to define a general coulombic lambda. This lambda
            only affects coulombic parameters of both molecules. It needs
            to be contained between 0 and 1. It has precedence over
            fep_lambda. Default is None
        coul1_lambda : float
            The value to define a coulombic lambda for exclusive atoms
            of molecule 1. This lambda only affects coulombic parameters
            of exclusive atoms of molecule 1. It needs to be contained
            between 0 and 1. It has precedence over coul_lambda or
            fep_lambda. Default is None
        coul2_lambda : float
            The value to define a coulombic lambda for exclusive atoms
            of molecule 2. This lambda only affects coulombic parameters
            of exclusive atoms of molecule 2. It needs to be contained
            between 0 and 1. It has precedence over coul_lambda or
            fep_lambda. Default is None
        vdw_lambda : float
            The value to define a vdw lambda. This lambda only
            affects van der Waals parameters. It needs to be contained
            between 0 and 1. It has precedence over fep_lambda.
            Default is None
        bonded_lambda : float
            The value to define a coulombic lambda. This lambda only
            affects bonded parameters. It needs to be contained
            between 0 and 1. It has precedence over fep_lambda.
            Default is None

        Returns
        -------
        alchemical_topology : a peleffy.topology.Topology
            The resulting alchemical topology
        """
        # Define lambdas
        fep_lambda = FEPLambda(fep_lambda)
        coul_lambda = CoulombicLambda(coul_lambda)
        coul1_lambda = Coulombic1Lambda(coul1_lambda)
        coul2_lambda = Coulombic2Lambda(coul2_lambda)
        vdw_lambda = VanDerWaalsLambda(vdw_lambda)
        bonded_lambda = BondedLambda(bonded_lambda)

        lambda_set = LambdaSet(fep_lambda, coul_lambda, coul1_lambda,
                               coul2_lambda, vdw_lambda, bonded_lambda)

        alchemical_topology = self.topology_from_lambda_set(lambda_set)

        return alchemical_topology

    def topology_from_lambda_set(self, lambda_set):
        """
        Given a lambda, it returns an alchemical topology after
        combining both input topologies.

        Parameters
        ----------
        lambda_set : a peleffy.topology.alchemy.LambdaSet object
            The set of lambdas to use in the generation of the
            alchemical topology

        Returns
        -------
        alchemical_topology : a peleffy.topology.Topology
            The resulting alchemical topology
        """
        from copy import deepcopy

        alchemical_topology = deepcopy(self._joint_topology)

        # Define mappers
        mol1_mapped_atoms = [atom_pair[0] for atom_pair in self.mapping]
        mol2_mapped_atoms = [atom_pair[1] for atom_pair in self.mapping]
        mol1_to_mol2_map = dict(zip(mol1_mapped_atoms, mol2_mapped_atoms))

        for atom_idx, atom in enumerate(alchemical_topology.atoms):
            if atom_idx in self._exclusive_atoms:
                atom.apply_lambda(["sigma", "epsilon", "born_radius",
                                   "SASA_radius", "nonpolar_gamma",
                                   "nonpolar_alpha"],
                                  lambda_set.get_lambda_for_vdw(),
                                  reverse=False)
                atom.apply_lambda(["charge"],
                                  lambda_set.get_lambda_for_coulomb1(),
                                  reverse=False)

            if atom_idx in self._non_native_atoms:
                atom.apply_lambda(["sigma", "epsilon", "born_radius",
                                   "SASA_radius", "nonpolar_gamma",
                                   "nonpolar_alpha"],
                                  lambda_set.get_lambda_for_vdw(),
                                  reverse=True)
                atom.apply_lambda(["charge"],
                                  lambda_set.get_lambda_for_coulomb2(),
                                  reverse=True)

            if atom_idx in mol1_mapped_atoms:
                mol2_idx = mol1_to_mol2_map[atom_idx]
                mol2_atom = self.topology2.atoms[mol2_idx]
                atom.apply_lambda(["sigma", "epsilon", "born_radius",
                                   "SASA_radius", "nonpolar_gamma",
                                   "nonpolar_alpha"],
                                  lambda_set.get_lambda_for_vdw(),
                                  reverse=False,
                                  final_state=mol2_atom)
                atom.apply_lambda(["charge"],
                                  lambda_set.get_lambda_for_coulomb(),
                                  reverse=False,
                                  final_state=mol2_atom)

        for bond_idx, bond in enumerate(alchemical_topology.bonds):
            if bond_idx in self._exclusive_bonds:
                bond.apply_lambda(["spring_constant", "eq_dist"],
                                  lambda_set.get_lambda_for_bonded(),
                                  reverse=False)

            if bond_idx in self._non_native_bonds:
                bond.apply_lambda(["spring_constant", "eq_dist"],
                                  lambda_set.get_lambda_for_bonded(),
                                  reverse=True)

            atom1_idx = bond.atom1_idx
            atom2_idx = bond.atom2_idx
            if (atom1_idx in mol1_mapped_atoms and
                    atom2_idx in mol1_mapped_atoms):
                mol_ids = (mol1_to_mol2_map[atom1_idx],
                           mol1_to_mol2_map[atom2_idx])

                for mol2_bond in self.topology2.bonds:
                    if (mol2_bond.atom1_idx in mol_ids and
                            mol2_bond.atom2_idx in mol_ids):
                        bond.apply_lambda(["spring_constant", "eq_dist"],
                                          lambda_set.get_lambda_for_bonded(),
                                          reverse=False,
                                          final_state=mol2_bond)

        for angle_idx, angle in enumerate(alchemical_topology.angles):
            if angle_idx in self._exclusive_angles:
                angle.apply_lambda(["spring_constant", "eq_angle"],
                                   lambda_set.get_lambda_for_bonded(),
                                   reverse=False)

            if angle_idx in self._non_native_angles:
                angle.apply_lambda(["spring_constant", "eq_angle"],
                                   lambda_set.get_lambda_for_bonded(),
                                   reverse=True)

            atom1_idx = angle.atom1_idx
            atom2_idx = angle.atom2_idx
            atom3_idx = angle.atom3_idx
            if (atom1_idx in mol1_mapped_atoms and
                atom2_idx in mol1_mapped_atoms and
                    atom3_idx in mol1_mapped_atoms):
                mol_ids = (mol1_to_mol2_map[atom1_idx],
                           mol1_to_mol2_map[atom2_idx],
                           mol1_to_mol2_map[atom3_idx])

                for mol2_angle in self.topology2.angles:
                    if (mol2_angle.atom1_idx in mol_ids and
                        mol2_angle.atom2_idx in mol_ids and
                            mol2_angle.atom3_idx in mol_ids):
                        angle.apply_lambda(["spring_constant", "eq_angle"],
                                           lambda_set.get_lambda_for_bonded(),
                                           reverse=False,
                                           final_state=mol2_angle)

        # Joint topology cannot have mutual propers
        for proper_idx, proper in enumerate(alchemical_topology.propers):
            if proper_idx in self._exclusive_propers:
                proper.apply_lambda(["constant"],
                                    lambda_set.get_lambda_for_bonded(),
                                    reverse=False)

            if proper_idx in self._non_native_propers:
                proper.apply_lambda(["constant"],
                                    lambda_set.get_lambda_for_bonded(),
                                    reverse=True)

        # Joint topology cannot have mutual impropers
        for improper_idx, improper in enumerate(alchemical_topology.impropers):
            if improper_idx in self._exclusive_impropers:
                improper.apply_lambda(["constant"],
                                      lambda_set.get_lambda_for_bonded(),
                                      reverse=False)

            if improper_idx in self._non_native_impropers:
                improper.apply_lambda(["constant"],
                                      lambda_set.get_lambda_for_bonded(),
                                      reverse=True)

        return alchemical_topology

    def _join_topologies(self):
        """
        It joins the both topologies into a single one that contains
        all topological elements.

        Returns
        -------
        joint_topology : a peleffy.topology.Topology
            The resulting alchemical topology
        non_native_atoms : list[int]
            The list of atom indices that were added to topology 1
        non_native_bonds : list[int]
            The list of bond indices that were added to topology 1
        non_native_angles : list[int]
            The list of angle indices that were added to topology 1
        non_native_propers : list[int]
            The list of proper indices that were added to topology 1
        non_native_impropers : list[int]
            The list of improper indices that were added to topology 1
        mol2_to_alc_map : dict[int, int]]
            The dictionary that pairs molecule 2 indices with alchemical
            molecule indices
        """
        from copy import deepcopy
        from peleffy.topology import Topology
        from peleffy.utils.toolkits import RDKitToolkitWrapper
        from peleffy.utils import Logger

        # General initializers
        logger = Logger()

        # First initialize the joint topology with topology 1
        joint_topology = deepcopy(self.topology1)

        # Change molecule tag
        joint_topology.molecule.set_tag('HYB')

        # Initialize list of non native topological elements
        non_native_atoms = list()
        non_native_bonds = list()
        non_native_angles = list()
        non_native_propers = list()
        non_native_impropers = list()

        # Define mappers
        mol1_mapped_atoms = [atom_pair[0] for atom_pair in self.mapping]
        mol2_mapped_atoms = [atom_pair[1] for atom_pair in self.mapping]
        mol2_to_mol1_map = dict(zip(mol2_mapped_atoms, mol1_mapped_atoms))

        # Add atoms from topology 2 that are missing in topology 1
        mol2_to_alc_map = dict()
        for atom_idx, atom in enumerate(self.topology2.atoms):
            if atom_idx not in mol2_mapped_atoms:
                new_atom = deepcopy(atom)

                # Handle index of new atom
                new_index = len(joint_topology.atoms)
                new_atom.set_index(new_index)
                mol2_to_alc_map[atom_idx] = new_index

                # Handle core allocation of new atom
                new_atom.set_as_branch()

                # Add new atom to the alchemical topology
                non_native_atoms.append(new_index)
                joint_topology.add_atom(new_atom)

            else:
                mol2_to_alc_map[atom_idx] = mol2_to_mol1_map[atom_idx]

        # Get an arbitrary absolute parent for molecule 2 (as long as it is in the MCS)
        absolute_parent = None
        for atom_idx in range(0, len(self.topology2.atoms)):
            if atom_idx in mol2_mapped_atoms:
                absolute_parent = atom_idx
                break
        else:
            logger.error(['Error: no atom in the MCS found in molecule ' +
                          f'{self.molecule1.name}'])

        # Get parent ids according to this new absolute parent of molecule 2
        mol2_parent_idxs = self.molecule2.graph.get_parents(absolute_parent)

        # Assign parents
        for atom_idx, atom in enumerate(self.topology2.atoms):
            if atom_idx not in mol2_mapped_atoms:
                alc_child_idx = mol2_to_alc_map[atom_idx]
                alc_parent_idx = mol2_to_alc_map[mol2_parent_idxs[atom_idx]]
                alc_parent_atom = joint_topology.atoms[alc_parent_idx]
                joint_topology.atoms[alc_child_idx].set_parent(
                    alc_parent_atom)

        # Add bonds
        for bond in self.topology2.bonds:
            atom1_idx = bond.atom1_idx
            atom2_idx = bond.atom2_idx

            if (atom1_idx not in mol2_mapped_atoms or
                    atom2_idx not in mol2_mapped_atoms):
                new_bond = deepcopy(bond)

                # Handle index of new atom
                new_index = len(joint_topology.bonds)
                new_bond.set_index(new_index)

                # Handle atom indices
                new_bond.set_atom1_idx(mol2_to_alc_map[atom1_idx])
                new_bond.set_atom2_idx(mol2_to_alc_map[atom2_idx])

                # Add new bond to the alchemical topology
                non_native_bonds.append(new_index)
                joint_topology.add_bond(new_bond)

        # Add angles
        for angle in self.topology2.angles:
            atom1_idx = angle.atom1_idx
            atom2_idx = angle.atom2_idx
            atom3_idx = angle.atom3_idx

            if (atom1_idx not in mol2_mapped_atoms or
                atom2_idx not in mol2_mapped_atoms or
                    atom3_idx not in mol2_mapped_atoms):
                new_angle = deepcopy(angle)

                # Handle index of new atom
                new_index = len(joint_topology.angles)
                new_angle.set_index(new_index)

                # Handle atom indices
                new_angle.set_atom1_idx(mol2_to_alc_map[atom1_idx])
                new_angle.set_atom2_idx(mol2_to_alc_map[atom2_idx])
                new_angle.set_atom3_idx(mol2_to_alc_map[atom3_idx])

                # Add new angle to the alchemical topology
                non_native_angles.append(new_index)
                joint_topology.add_angle(new_angle)

        # Add propers
        for proper in self.topology2.propers:
            atom1_idx = proper.atom1_idx
            atom2_idx = proper.atom2_idx
            atom3_idx = proper.atom3_idx
            atom4_idx = proper.atom1_idx

            # Add a copy of the proper
            new_proper = deepcopy(proper)

            # Handle index of new atom
            new_index = len(joint_topology.propers)
            new_proper.set_index(new_index)

            # Handle atom indices
            new_proper.set_atom1_idx(mol2_to_alc_map[atom1_idx])
            new_proper.set_atom2_idx(mol2_to_alc_map[atom2_idx])
            new_proper.set_atom3_idx(mol2_to_alc_map[atom3_idx])
            new_proper.set_atom4_idx(mol2_to_alc_map[atom4_idx])

            # Add new proper to the alchemical topology
            non_native_propers.append(new_index)
            joint_topology.add_proper(new_proper)

        # Add impropers
        for improper in self.topology2.impropers:
            atom1_idx = improper.atom1_idx
            atom2_idx = improper.atom2_idx
            atom3_idx = improper.atom3_idx
            atom4_idx = improper.atom1_idx

            # Add a copy of the improper
            new_improper = deepcopy(improper)

            # Handle index of new atom
            new_index = len(joint_topology.impropers)
            new_improper.set_index(new_index)

            # Handle atom indices
            new_improper.set_atom1_idx(mol2_to_alc_map[atom1_idx])
            new_improper.set_atom2_idx(mol2_to_alc_map[atom2_idx])
            new_improper.set_atom3_idx(mol2_to_alc_map[atom3_idx])
            new_improper.set_atom4_idx(mol2_to_alc_map[atom4_idx])

            # Add new improper to the alchemical topology
            non_native_impropers.append(new_index)
            joint_topology.add_improper(new_improper)

        return joint_topology, non_native_atoms, non_native_bonds, \
            non_native_angles, non_native_propers, non_native_impropers, \
            mol2_to_alc_map

    def _get_exclusive_elements(self):
        """
        It identifies those topological elements that are exclusive
        of topology 1. The condition to be exclusive is to belong to
        topology 1 but not to the MCS.

        Returns
        -------
        exclusive_atoms : list[int]
            The list of atom indices that are exclusive of topology 1
        exclusive_bonds : list[int]
            The list of bond indices that are exclusive of topology 1
        exclusive_angles : list[int]
            The list of angle indices that are exclusive of topology 1
        exclusive_propers : list[int]
            The list of proper indices that are exclusive of topology 1
        exclusive_impropers : list[int]
            The list of improper indices that are exclusive of topology 1
        """
        # Define mappers
        mol1_mapped_atoms = [atom_pair[0] for atom_pair in self.mapping]

        # Initialize list of exclusive topological elements
        exclusive_atoms = list()
        exclusive_bonds = list()
        exclusive_angles = list()
        exclusive_propers = list()
        exclusive_impropers = list()

        # Identify atoms
        for atom_idx, atom in enumerate(self.topology1.atoms):
            if atom_idx not in mol1_mapped_atoms:
                exclusive_atoms.append(atom_idx)

        # Identify bonds
        for bond_idx, bond in enumerate(self.topology1.bonds):
            atom1_idx = bond.atom1_idx
            atom2_idx = bond.atom2_idx

            if (atom1_idx not in mol1_mapped_atoms or
                    atom2_idx not in mol1_mapped_atoms):
                exclusive_bonds.append(bond_idx)

        # Identify angles
        for angle_idx, angle in enumerate(self.topology1.angles):
            atom1_idx = angle.atom1_idx
            atom2_idx = angle.atom2_idx
            atom3_idx = angle.atom3_idx

            if (atom1_idx not in mol1_mapped_atoms or
                atom2_idx not in mol1_mapped_atoms or
                    atom3_idx not in mol1_mapped_atoms):
                exclusive_angles.append(angle_idx)

       # Identify propers
        for proper_idx, proper in enumerate(self.topology1.propers):
            exclusive_propers.append(proper_idx)

       # Identify impropers
        for improper_idx, improper in enumerate(self.topology1.impropers):
            exclusive_impropers.append(improper_idx)

        return exclusive_atoms, exclusive_bonds, exclusive_angles, \
               exclusive_propers, exclusive_impropers

    def _find_connections(self):
        """
        It finds the connections between molecule 1 and molecule2.

        Returns
        -------
        connections : list[tuple[int, int]]
            The list of connections between molecule 1 and non
            native atoms of molecule 2
        """
        # TODO move to rdkit toolkit
        mol2_mapped_atoms = [atom_pair[1] for atom_pair in self.mapping]

        connections = list()
        for atom in self.molecule2.rdkit_molecule.GetAtoms():
            if atom.GetIdx() in mol2_mapped_atoms:
                for bond in atom.GetBonds():
                    index1 = self._mol2_to_alc_map[bond.GetBeginAtomIdx()]
                    index2 = self._mol2_to_alc_map[bond.GetEndAtomIdx()]
                    if index1 in self._non_native_atoms:
                        connections.append((index1, index2))
                    elif index2 in self._non_native_atoms:
                        connections.append((index1, index2))

        return connections

    def _generate_alchemical_graph(self):
        """
        If generates the alchemical graph and rotamers.

        Returns
        -------
        alchemical_graph : a peleffy.topology.rotamer.MolecularGraph object
            The molecular graph containing the alchemical structure
        rotamers : list[list]
            The list of rotamers grouped by the branch they belong to
        """
        from copy import deepcopy

        # Handle peleffy Logger
        from peleffy.utils import Logger

        logger = Logger()
        
        # Define mappers
        mol1_mapped_atoms = [atom_pair[0] for atom_pair in self.mapping]
        mol2_mapped_atoms = [atom_pair[1] for atom_pair in self.mapping]
        mol2_to_mol1_map = dict(zip(mol2_mapped_atoms, mol1_mapped_atoms))

        # Copy graph of molecule 1
        alchemical_graph = deepcopy(self.molecule1._graph)

        # Fix conflicts on common edges of both molecules
        for mol2_edge in self.molecule2.graph.edges:
            if (mol2_edge[0] in mol2_to_mol1_map.keys() and
                    mol2_edge[1] in mol2_to_mol1_map.keys()):
                # Get indices of both atoms of this edge
                index1 = mol2_to_mol1_map[mol2_edge[0]]
                index2 = mol2_to_mol1_map[mol2_edge[1]]

                # Get weights in each graph
                # Maybe the edge is not defined yet in the alchemical graph
                if index2 not in alchemical_graph[index1]:
                    weight1 = 0
                    alchemical_graph.add_edge(index1, index2,
                                              weight=weight1)
                else:
                    weight1 = alchemical_graph[index1][index2]['weight']

                weight2 = \
                    self.molecule2.graph[mol2_edge[0]][mol2_edge[1]]['weight']

                # We keep the higher weight, so in case that a bond is
                # rotatable in one molecule but not in the order
                # one, it will be defined as rotatable
                weight = int(max((weight1, weight2)))
                alchemical_graph[index1][index2]['weight'] = weight

                # Update nrot_neighbors list
                node1 = alchemical_graph.nodes[index1]
                node2 = alchemical_graph.nodes[index2]
                if weight == 1:  # Rotatable
                    if index2 in node1['nrot_neighbors']:
                        node1['nrot_neighbors'].remove(index2)
                    if index1 in node2['nrot_neighbors']:
                        node2['nrot_neighbors'].remove(index1)
                else:  # Non rotatable
                    if index2 not in node1['nrot_neighbors']:
                        node1['nrot_neighbors'].append(index2)
                    if index1 not in node2['nrot_neighbors']:
                        node2['nrot_neighbors'].append(index1)

        # Add non native nodes
        for mol2_node1 in self.molecule2.graph.nodes:
            index1 = self._mol2_to_alc_map[mol2_node1]
            if index1 in self._non_native_atoms:
                name = self.molecule2.graph.nodes[mol2_node1]['pdb_name']
                nrot_neighbors = \
                    self.molecule2.graph.nodes[mol2_node1]['nrot_neighbors']
                nrot_neighbors = [self._mol2_to_alc_map[neighbor]
                                  for neighbor in nrot_neighbors]
                alchemical_graph.add_node(index1, pdb_name=name,
                                          nrot_neighbors=nrot_neighbors)

                for mol2_node2 in self.molecule2.graph[mol2_node1]:
                    index2 = self._mol2_to_alc_map[mol2_node2]
                    if not alchemical_graph.has_edge(index1, index2):
                        rotatable = \
                            self.molecule2.graph[mol2_node1][mol2_node2]['weight']
                        alchemical_graph.add_edge(index1, index2,
                                                  weight=int(rotatable))

        alchemical_graph._build_core_nodes()

        rotamers = alchemical_graph.get_rotamers()

        # Update core/branch location
        for atom in self._joint_topology.atoms:
            if atom.index in alchemical_graph.core_nodes:
                atom.set_as_core()
            else:
                atom.set_as_branch()

        # Find absolute parent atom
        absolute_parent = None
        for atom in self._joint_topology.atoms:
            if atom.core:
                absolute_parent = atom.index
                break
        else:
            logger.error(['Error: no core atom found in hybrid molecule'])

        # Get parent indexes from the molecular graph
        parent_idxs = alchemical_graph.get_parents(absolute_parent)

        # Assert parent_idxs has right length
        if len(parent_idxs) != len(self._joint_topology.atoms):
            logger.error(['Error: invalid number of parents obtained for ' +
                          'the hybrid molecule'])

        # Assign parent atoms
        for atom in self._joint_topology.atoms:
            parent_idx = parent_idxs[atom.index]
            if parent_idx is not None:
                atom.set_parent(self._joint_topology.atoms[parent_idx])
            else:
                atom.set_parent(None)

        return alchemical_graph, rotamers

    def _assign_pdb_atom_names(self):
        """
        It assigns consistent PDB atom names to the alchemical molecule.
        """
        from peleffy.topology import Molecule
        from peleffy.utils.toolkits import RDKitToolkitWrapper

        rdkit_wrapper = RDKitToolkitWrapper()

        # Combine molecules
        mol_combo = \
            rdkit_wrapper.alchemical_combination(self.molecule1.rdkit_molecule,
                                                 self.molecule2.rdkit_molecule,
                                                 self.mapping,
                                                 self.connections)

        # Generate a dummy peleffy Molecule with the required information
        # to extract PDB atom names
        molecule = Molecule()
        molecule._rdkit_molecule = mol_combo
        molecule.set_tag('HYB')

        # Extract PDB atom names
        atom_names = rdkit_wrapper.get_atom_names(molecule)

        # Assign PDB atom names
        for atom_idx, atom in enumerate(self._joint_topology.atoms):
            atom.set_PDB_name(atom_names[atom_idx].replace(' ', '_'))

    def hybrid_to_pdb(self, path):
        """
        Writes the alchemical molecule to a PDB file.

        Parameters
        ----------
        path : str
            The path where to save the PDB file
        """
        from copy import deepcopy
        from peleffy.topology import Molecule

        from peleffy.utils.toolkits import RDKitToolkitWrapper

        rdkit_wrapper = RDKitToolkitWrapper()

        # Combine molecules
        mol_combo = \
            rdkit_wrapper.alchemical_combination(self.molecule1.rdkit_molecule,
                                                 self.molecule2.rdkit_molecule,
                                                 self.mapping,
                                                 self.connections)

        # Generate a dummy peleffy Molecule with the required information
        # to write it as a PDB file
        molecule = Molecule()
        molecule._rdkit_molecule = mol_combo
        molecule.set_tag('HYB')

        rdkit_wrapper.to_pdb_file(molecule, path)

    def molecule1_to_pdb(self, path):
        """
        Writes the first molecule of the Alchemizer representation
        to a PDB file.

        Parameters
        ----------
        path : str
            The path where to save the PDB file
        """
        # Write it directly
        self.molecule1.to_pdb_file(path)

    def molecule2_to_pdb(self, path):
        """
        Writes the first molecule of the Alchemizer representation
        to a PDB file.

        Parameters
        ----------
        path : str
            The path where to save the PDB file
        """
        from peleffy.topology import Molecule

        # Align it to molecule 1
        from peleffy.utils.toolkits import RDKitToolkitWrapper

        rdkit_wrapper = RDKitToolkitWrapper()
        mol2_aligned = \
            rdkit_wrapper.align_molecules(self.molecule1.rdkit_molecule,
                                          self.molecule2.rdkit_molecule,
                                          self.mapping)

        # Generate a dummy peleffy Molecule with the required information
        # to write it as a PDB file
        molecule = Molecule()
        molecule._rdkit_molecule = mol2_aligned
        molecule.set_tag(self.molecule2.tag)

        # Write it
        rdkit_wrapper.to_pdb_file(molecule, path)

    def rotamer_library_to_file(self, path, fep_lambda=None,
                                coul_lambda=None, coul1_lambda=None,
                                coul2_lambda=None, vdw_lambda=None,
                                bonded_lambda=None):
        """
        It saves the alchemical rotamer library, which is the combination
        of the rotamer libraries of both molecules, to the path that
        is supplied.

        Parameters
        ----------
        path : str
            The path where to save the rotamer library
        fep_lambda : float
            The value to define an FEP lambda. This lambda affects
            all the parameters. It needs to be contained between
            0 and 1. Default is None
        coul_lambda : float
            The value to define a general coulombic lambda. This lambda
            only affects coulombic parameters of both molecules. It needs
            to be contained between 0 and 1. It has precedence over
            fep_lambda. Default is None
        coul1_lambda : float
            The value to define a coulombic lambda for exclusive atoms
            of molecule 1. This lambda only affects coulombic parameters
            of exclusive atoms of molecule 1. It needs to be contained
            between 0 and 1. It has precedence over coul_lambda or
            fep_lambda. Default is None
        coul2_lambda : float
            The value to define a coulombic lambda for exclusive atoms
            of molecule 2. This lambda only affects coulombic parameters
            of exclusive atoms of molecule 2. It needs to be contained
            between 0 and 1. It has precedence over coul_lambda or
            fep_lambda. Default is None
        vdw_lambda : float
            The value to define a vdw lambda. This lambda only
            affects van der Waals parameters. It needs to be contained
            between 0 and 1. It has precedence over fep_lambda.
            Default is None
        bonded_lambda : float
            The value to define a coulombic lambda. This lambda only
            affects bonded parameters. It needs to be contained
            between 0 and 1. It has precedence over fep_lambda.
            Default is None
        """

        at_least_one = fep_lambda is not None or \
            coul_lambda is not None or coul1_lambda is not None or \
            coul2_lambda is not None or vdw_lambda is not None or \
            bonded_lambda is not None

        # Define lambdas
        fep_lambda = FEPLambda(fep_lambda)
        coul_lambda = CoulombicLambda(coul_lambda)
        coul1_lambda = Coulombic1Lambda(coul1_lambda)
        coul2_lambda = Coulombic2Lambda(coul2_lambda)
        vdw_lambda = VanDerWaalsLambda(vdw_lambda)
        bonded_lambda = BondedLambda(bonded_lambda)

        lambda_set = LambdaSet(fep_lambda, coul_lambda, coul1_lambda,
                               coul2_lambda, vdw_lambda, bonded_lambda)

        if (at_least_one and
            lambda_set.get_lambda_for_bonded() == 0.0 and
            lambda_set.get_lambda_for_vdw() == 0.0 and
            lambda_set.get_lambda_for_coulomb() == 0.0 and
            lambda_set.get_lambda_for_coulomb1() == 0.0 and
                lambda_set.get_lambda_for_coulomb2() == 0.0):
            rotamers = self.molecule1.rotamers
            mapping = False

        elif (at_least_one and
              lambda_set.get_lambda_for_bonded() == 1.0 and
              lambda_set.get_lambda_for_vdw() == 1.0 and
              lambda_set.get_lambda_for_coulomb() == 1.0 and
              lambda_set.get_lambda_for_coulomb1() == 1.0 and
                  lambda_set.get_lambda_for_coulomb2() == 1.0):
            rotamers = self.molecule2.rotamers
            mapping = True

        else:
            rotamers = self._rotamers
            mapping = False

        # Initial definitions
        pdb_atom_names = [atom.PDB_name.replace(' ', '_',)
                          for atom in self._joint_topology.atoms]
        molecule_tag = self._joint_topology.molecule.tag

        with open(path, 'w') as file:
            file.write('rot assign res {} &\n'.format(molecule_tag))
            for i, rotamer_branches in enumerate(rotamers):
                if i > 0:
                    file.write('     newgrp &\n')
                for rotamer in rotamer_branches:
                    index1 = rotamer.index1
                    index2 = rotamer.index2

                    if mapping:
                        index1 = self._mol2_to_alc_map[index1]
                        index2 = self._mol2_to_alc_map[index2]

                    atom_name1 = pdb_atom_names[index1]
                    atom_name2 = pdb_atom_names[index2]
                    file.write('   sidelib FREE{} {} {} &\n'.format(
                        rotamer.resolution, atom_name1, atom_name2))

    def obc_parameters_to_file(self, path, fep_lambda=None,
                               coul_lambda=None, coul1_lambda=None,
                               coul2_lambda=None, vdw_lambda=None,
                               bonded_lambda=None):
        """
        It saves the alchemical OBC parameters, which is the combination
        of the OBC parameters of both molecules, to the path that
        is supplied.

        Parameters
        ----------
        path : str
            The path where to save the OBC parameters template
        fep_lambda : float
            The value to define an FEP lambda. This lambda affects
            all the parameters. It needs to be contained between
            0 and 1. Default is None
        coul_lambda : float
            The value to define a general coulombic lambda. This lambda
            only affects coulombic parameters of both molecules. It needs
            to be contained between 0 and 1. It has precedence over
            fep_lambda. Default is None
        coul1_lambda : float
            The value to define a coulombic lambda for exclusive atoms
            of molecule 1. This lambda only affects coulombic parameters
            of exclusive atoms of molecule 1. It needs to be contained
            between 0 and 1. It has precedence over coul_lambda or
            fep_lambda. Default is None
        coul2_lambda : float
            The value to define a coulombic lambda for exclusive atoms
            of molecule 2. This lambda only affects coulombic parameters
            of exclusive atoms of molecule 2. It needs to be contained
            between 0 and 1. It has precedence over coul_lambda or
            fep_lambda. Default is None
        vdw_lambda : float
            The value to define a vdw lambda. This lambda only
            affects van der Waals parameters. It needs to be contained
            between 0 and 1. It has precedence over fep_lambda.
            Default is None
        bonded_lambda : float
            The value to define a coulombic lambda. This lambda only
            affects bonded parameters. It needs to be contained
            between 0 and 1. It has precedence over fep_lambda.
            Default is None

        Returns
        -------
        path : str
            The path where to save the rotamer library
        """

        # Handle peleffy Logger
        from peleffy.utils import Logger

        logger = Logger()
        log_level = logger.get_level()
        logger.set_level('ERROR')

        # Define lambdas
        fep_lambda = FEPLambda(fep_lambda)
        coul_lambda = CoulombicLambda(coul_lambda)
        coul1_lambda = Coulombic1Lambda(coul1_lambda)
        coul2_lambda = Coulombic2Lambda(coul2_lambda)
        vdw_lambda = VanDerWaalsLambda(vdw_lambda)
        bonded_lambda = BondedLambda(bonded_lambda)

        lambda_set = LambdaSet(fep_lambda, coul_lambda, coul1_lambda,
                               coul2_lambda, vdw_lambda, bonded_lambda)

        # Define mappers
        mol1_mapped_atoms = [atom_pair[0] for atom_pair in self.mapping]
        mol2_mapped_atoms = [atom_pair[1] for atom_pair in self.mapping]
        mol1_to_mol2_map = dict(zip(mol1_mapped_atoms, mol2_mapped_atoms))

        # Generate individual OBC parameters
        from copy import deepcopy
        from peleffy.solvent import OBC2

        mol1_obc_params = OBC2(self.topology1)
        mol2_obc_params = OBC2(self.topology2)

        # Generate alchemical OBC parameters object
        alchemical_obc_params = deepcopy(mol1_obc_params)
        alchemical_obc_params._topologies = [self._joint_topology, ]

        # Get OBC parameters of molecule 1
        radii1 = alchemical_obc_params._radii[0]
        scales1 = alchemical_obc_params._scales[0]
        radii2 = mol2_obc_params._radii[0]
        scales2 = mol2_obc_params._scales[0]

        for atom_idx, atom in enumerate(self._joint_topology.atoms):
            if atom_idx in self._exclusive_atoms:
                lambda_value = 1.0 - lambda_set.get_lambda_for_coulomb1()
                radius = radii1[(atom_idx, )] * lambda_value
                scale = scales1[(atom_idx, )] * lambda_value

            elif atom_idx in self._non_native_atoms:
                for mol2_index, alc_index in self._mol2_to_alc_map.items():
                    if alc_index == atom_idx:
                        lambda_value = lambda_set.get_lambda_for_coulomb2()
                        radius = radii2[(mol2_index, )] * lambda_value
                        scale = scales2[(mol2_index, )] * lambda_value
                        break
                else:
                    logger.error(['Error: mapping for atom index ' +
                                  f'{atom_idx} not found in the ' +
                                  'hybrid molecule'])
                    radius = 0
                    scale = 0

            elif atom_idx in mol1_mapped_atoms:
                mol2_idx = mol1_to_mol2_map[atom_idx]
                radius2 = mol2_obc_params._radii[0][(mol2_idx, )]
                scale2 = mol2_obc_params._scales[0][(mol2_idx, )]

                lambda_value = 1.0 - lambda_set.get_lambda_for_coulomb2()
                radius = radii1[(atom_idx, )] * lambda_value \
                    + (1.0 - lambda_value) * radius2
                scale = scales1[(atom_idx, )] * lambda_value \
                    + (1.0 - lambda_value) * scale2

            alchemical_obc_params._radii[0][(atom_idx, )] = radius
            alchemical_obc_params._scales[0][(atom_idx, )] = scale

        alchemical_obc_params.to_file(path)

        logger.set_level(log_level)

    def _ipython_display_(self):
        """
        It returns a representation of the alchemical mapping.

        Returns
        -------
        mapping_representation : an IPython display object
            Displayable RDKit molecules with mapping information
        """
        from IPython.display import display

        return display(self._mapper)


class Lambda(ABC):
    """
    It defines the Lambda class.
    """

    _TYPE = ""

    def __init__(self, value=None):
        """
        It initializes a Lambda object.

        Parameters
        ----------
        value : float
            The value of this Lambda object. It needs to be
            contained between 0 and 1. Default is None
        """
        if value is not None:
            try:
                value = float(value)
            except ValueError:
                raise ValueError("Invalid value for a lambda: " +
                                 f"\'{value}\'")
            if (value > 1) or (value < 0):
                raise ValueError("Invalid value for a lambda: " +
                                 f"\'{value}\'. " +
                                 "It has to be between 0 and 1")

        self._value = value

    @property
    def value(self):
        """
        It returns the value of this Lambda object.

        Returns
        -------
        value : float or None
            The value of this Lambda object. It can be None if the
            value for this Lambda object has not been set
        """
        return self._value

    @property
    def type(self):
        """
        It returns the type of this Lambda object.

        Returns
        -------
        type : str
            The type of this Lambda object
        """
        return self._TYPE

    @property
    def is_set(self):
        """
        It answers whether the value of this Lambda object has
        been set or not.

        Returns
        -------
        is_set : bool
            It is true only if the value of Lambda object has been
            set
        """
        return self.value is not None


class FEPLambda(Lambda):
    """
    It defines the FEPLambda class. It affects all parameters.
    """
    _TYPE = "fep"

class CoulombicLambda(Lambda):
    """
    It defines the CoulombicLambda class. It affects only coulombic
    parameters involving both molecules.
    """
    _TYPE = "coulombic"


class Coulombic1Lambda(Lambda):
    """
    It defines the CoulombicLambda1 class. It affects only coulombic
    parameters involving exclusive atoms of molecule 1.
    """
    _TYPE = "coulombic1"


class Coulombic2Lambda(Lambda):
    """
    It defines the CoulombicLambda2 class. It affects only coulombic
    parameters involving exclusive atoms of molecule 2.
    """
    _TYPE = "coulombic2"


class VanDerWaalsLambda(Lambda):
    """
    It defines the VanDerWaalsLambda class. It affects only van der Waals
    parameters.
    """
    _TYPE = "vdw"


class BondedLambda(Lambda):
    """
    It defines the BondedLambda class. It affects only bonded parameters.
    """
    _TYPE = "bonded"


class LambdaSet(object):
    """
    It defines the LambdaSet class.
    """

    def __init__(self, fep_lambda, coul_lambda, coul1_lambda, coul2_lambda,
                 vdw_lambda, bonded_lambda):
        """
        It initializes a LambdaSet object which stores all the different
        types of lambda.

        Parameters
        ----------
        fep_lambda : a peleffy.topology.alchemy.FEPLambda object
            The fep lambda
        coul_lambda : a peleffy.topology.alchemy.CoulombicLambda object
            The coulombic lambda for both molecules
        coul1_lambda : a peleffy.topology.alchemy.Coulombic1Lambda object
            The coulombic lambda for exclusive atoms of molecule 1
        coul2_lambda : a peleffy.topology.alchemy.Coulombic2Lambda object
            The coulombic lambda for exclusive atoms of molecule 2
        vdw_lambda : a peleffy.topology.alchemy.VanDerWaalsLambda object
            The van der Waals lambda
        bonded_lambda : a peleffy.topology.alchemy.BondedLambda object
            The bonded lambda
        """
        import peleffy

        # Check parameters
        if not isinstance(fep_lambda,
                          peleffy.topology.alchemistry.FEPLambda):
            raise TypeError('Invalid fep_lambda supplied to LambdaSet')
        if not isinstance(coul_lambda,
                          peleffy.topology.alchemistry.CoulombicLambda):
            raise TypeError('Invalid coul_lambda supplied to LambdaSet')
        if not isinstance(coul1_lambda,
                          peleffy.topology.alchemistry.Coulombic1Lambda):
            raise TypeError('Invalid coul1_lambda supplied to LambdaSet')
        if not isinstance(coul2_lambda,
                          peleffy.topology.alchemistry.Coulombic2Lambda):
            raise TypeError('Invalid coul2_lambda supplied to LambdaSet')
        if not isinstance(vdw_lambda,
                          peleffy.topology.alchemistry.VanDerWaalsLambda):
            raise TypeError('Invalid vdw_lambda supplied to LambdaSet')
        if not isinstance(bonded_lambda,
                          peleffy.topology.alchemistry.BondedLambda):
            raise TypeError('Invalid bonded_lambda supplied to LambdaSet')

        self._fep_lambda = fep_lambda
        self._coul_lambda = coul_lambda
        self._coul1_lambda = coul1_lambda
        self._coul2_lambda = coul2_lambda
        self._vdw_lambda = vdw_lambda
        self._bonded_lambda = bonded_lambda

    @property
    def fep_lambda(self):
        """
        It returns the fep_lambda value.

        Returns
        -------
        fep_lambda : float
            The value of the fep_lambda
        """
        return self._fep_lambda

    @property
    def coul_lambda(self):
        """
        It returns the coul_lambda value.

        Returns
        -------
        coul_lambda : float
            The value of the coul_lambda
        """
        return self._coul_lambda

    @property
    def coul1_lambda(self):
        """
        It returns the coul1_lambda value.

        Returns
        -------
        coul1_lambda : float
            The value of the coul1_lambda
        """
        return self._coul1_lambda

    @property
    def coul2_lambda(self):
        """
        It returns the coul2_lambda value.

        Returns
        -------
        coul2_lambda : float
            The value of the coul2_lambda
        """
        return self._coul2_lambda

    @property
    def vdw_lambda(self):
        """
        It returns the vdw_lambda value.

        Returns
        -------
        vdw_lambda : float
            The value of the vdw_lambda
        """
        return self._vdw_lambda

    @property
    def bonded_lambda(self):
        """
        It returns the bonded_lambda value.

        Returns
        -------
        bonded_lambda : float
            The value of the bonded_lambda
        """
        return self._bonded_lambda

    def get_lambda_for_vdw(self):
        """
        It returns the lambda to be applied on van der Waals parameters.

        Returns
        -------
        lambda_value : float
            The lambda value to be applied on van der Waals parameters
        """
        if self.vdw_lambda.is_set:
            lambda_value = self.vdw_lambda.value

        elif self.fep_lambda.is_set:
            lambda_value = self.fep_lambda.value

        else:
            lambda_value = 0.0

        return lambda_value

    def get_lambda_for_coulomb(self):
        """
        It returns the lambda to be applied on Coulomb parameters of
        both molecules.

        Returns
        -------
        lambda_value : float
            The lambda value to be applied on Coulomb parameters of
            both molecules
        """
        if self.coul_lambda.is_set:
            lambda_value = self.coul_lambda.value

        elif self.fep_lambda.is_set:
            lambda_value = self.fep_lambda.value

        else:
            lambda_value = 0.0

        return lambda_value

    def get_lambda_for_coulomb1(self):
        """
        It returns the lambda to be applied on Coulomb parameters of
        exclusive atoms of molecule 1.

        Returns
        -------
        lambda_value : float
            The lambda value to be applied on Coulomb parameters of
            exclusive atoms of molecule 1
        """
        if self.coul1_lambda.is_set:
            lambda_value = self.coul1_lambda.value

        elif self.coul_lambda.is_set:
            lambda_value = self.coul_lambda.value

        elif self.fep_lambda.is_set:
            lambda_value = self.fep_lambda.value

        else:
            lambda_value = 0.0

        return lambda_value

    def get_lambda_for_coulomb2(self):
        """
        It returns the lambda to be applied on Coulomb parameters of
        exclusive atoms of molecule 2.

        Returns
        -------
        lambda_value : float
            The lambda value to be applied on Coulomb parameters of
            exclusive atoms of molecule 2
        """
        if self.coul2_lambda.is_set:
            lambda_value = self.coul2_lambda.value

        elif self.coul_lambda.is_set:
            lambda_value = self.coul_lambda.value

        elif self.fep_lambda.is_set:
            lambda_value = self.fep_lambda.value

        else:
            lambda_value = 0.0

        return lambda_value

    def get_lambda_for_bonded(self):
        """
        It returns the lambda to be applied on bonded parameters.

        Returns
        -------
        lambda_value : float
            The lambda value to be applied on bonded parameters
        """
        if self.bonded_lambda.is_set:
            lambda_value = self.bonded_lambda.value

        elif self.fep_lambda.is_set:
            lambda_value = self.fep_lambda.value

        else:
            lambda_value = 0.0

        return lambda_value



"""
This module contains the tests to check peleffy's molecular mapper.
"""

import pytest


class TestMapper(object):
    """
    It wraps all tests that involve Mapper class.
    """

    def test_mapper_initializer(self):
        """
        It checks the initialization of the Mapper class.
        """
        from peleffy.topology import Molecule
        from peleffy.topology import Mapper

        mol1 = Molecule(smiles='c1ccccc1', hydrogens_are_explicit=False)
        mol2 = Molecule(smiles='c1ccccc1C', hydrogens_are_explicit=False)

        # Check initializer with only the two molecules
        mapper = Mapper(mol1, mol2)

        # Check initializer with only include_hydrogens parameter
        mapper = Mapper(mol1, mol2, include_hydrogens=False)

        # Check initializer with bad types
        with pytest.raises(TypeError):
            mapper = Mapper(mol1.rdkit_molecule, mol2)

        with pytest.raises(TypeError):
            mapper = Mapper(mol1, "mol2")

    def test_mapper_mapping(self):
        """
        It validates the mapping.
        """
        from peleffy.topology import Molecule
        from peleffy.topology import Mapper

        # First mapping checker
        mol1 = Molecule(smiles='c1ccccc1', hydrogens_are_explicit=False)
        mol2 = Molecule(smiles='c1ccccc1C', hydrogens_are_explicit=False)

        mapper = Mapper(mol1, mol2, include_hydrogens=False)
        mapping = mapper.get_mapping()

        assert mapping == [(0, 0), (1, 1), (2, 2), (3, 3),
                           (4, 4), (5, 5)], 'Unexpected mapping'

        # Second mapping checker
        mol1 = Molecule(smiles='c1(C)ccccc1C', hydrogens_are_explicit=False)
        mol2 = Molecule(smiles='c1c(C)cccc1C', hydrogens_are_explicit=False)

        mapper = Mapper(mol1, mol2, include_hydrogens=False)
        mapping = mapper.get_mapping()

        assert mapping == [(0, 1), (1, 2), (2, 0), (3, 6),
                           (4, 5), (5, 4), (6, 3)], 'Unexpected mapping'

        # Third mapping checker with hydrogens
        mol1 = Molecule(smiles='c1ccccc1', hydrogens_are_explicit=False)
        mol2 = Molecule(smiles='c1ccccc1C', hydrogens_are_explicit=False)

        mapper = Mapper(mol1, mol2, include_hydrogens=True)
        mapping = mapper.get_mapping()

        assert mapping == [(0, 0), (1, 1), (2, 2), (3, 3),
                           (4, 4), (5, 5), (11, 6), (10, 11),
                           (9, 10), (8, 9), (7, 8), (6, 7)], \
            'Unexpected mapping'

        # Fourth mapping checker with hydrogens
        mol1 = Molecule(smiles='c1(C)ccccc1C', hydrogens_are_explicit=False)
        mol2 = Molecule(smiles='c1c(C)cccc1C', hydrogens_are_explicit=False)

        mapper = Mapper(mol1, mol2, include_hydrogens=True)
        mapping = mapper.get_mapping()

        assert mapping == [(0, 1), (1, 2), (8, 9), (9, 10),
                           (10, 11), (2, 0), (3, 6), (4, 5),
                           (5, 4), (6, 3), (7, 12), (14, 13),
                           (13, 14), (12, 7), (11, 8)], 'Unexpected mapping'



"""
This module handles all classes and functions related with the mapping
of two molecular topologies.
"""


class Mapper(object):
    """
    It defines the Mapper class.
    """

    _TIMEOUT = 150  # Timeout to find the MCS, in seconds

    def __init__(self, molecule1, molecule2, include_hydrogens=True):
        """
        Given two molecules, it finds the maximum common substructure
        (MCS) and maps their atoms.

        Parameters
        ----------
        molecule1: a peleffy.topology.Molecule
            The first molecule to map
        molecule2: a peleffy.topology.Molecule
            The second molecule to map
        include_hydrogens: bool
            Whether to include hydrogen atoms in the mapping or not.
            Default is True
        """

        # Check parameters
        import peleffy

        if (not isinstance(molecule1, peleffy.topology.Molecule)
                and not
                isinstance(molecule1, peleffy.topology.molecule.Molecule)):
            raise TypeError('Invalid input molecule 1')

        if (not isinstance(molecule2, peleffy.topology.Molecule)
                and not
                isinstance(molecule2, peleffy.topology.molecule.Molecule)):
            raise TypeError('Invalid input molecule 2')

        if molecule1.rdkit_molecule is None:
            raise ValueError('Molecule 1 has not been initialized')

        if molecule2.rdkit_molecule is None:
            raise ValueError('Molecule 2 has not been initialized')

        self._molecule1 = molecule1
        self._molecule2 = molecule2
        self._include_hydrogens = include_hydrogens

    def get_mcs(self):
        """
        It returns the Maximum Common Substructure (MCS) between
        both molecules.

        Parameters
        ----------
        mcs_mol : an RDKit.molecule object
            The resulting MCS molecule
        """
        from peleffy.utils.toolkits import RDKitToolkitWrapper

        rdkit_toolkit = RDKitToolkitWrapper()

        mcs_mol = rdkit_toolkit.get_mcs(self.molecule1, self.molecule2,
                                        self._include_hydrogens,
                                        self._TIMEOUT)

        return mcs_mol

    def get_mapping(self):
        """
        It returns the mapping between both molecules.

        Returns
        -------
        mapping : list[tuple]
            The list of atom pairs between both molecules, represented
            with tuples
        """
        from peleffy.utils.toolkits import RDKitToolkitWrapper

        rdkit_toolkit = RDKitToolkitWrapper()

        mcs_mol = self.get_mcs()

        mapping = rdkit_toolkit.get_atom_mapping(self.molecule1,
                                                 self.molecule2,
                                                 mcs_mol,
                                                 self._include_hydrogens)

        return mapping

    @property
    def molecule1(self):
        """
        It returns the first molecule to map.

        Returns
        -------
        molecule1 : a peleffy.topology.Molecule
            The first molecule to map
        """
        return self._molecule1

    @property
    def molecule2(self):
        """
        It returns the second molecule to map.

        Returns
        -------
        molecule2 : a peleffy.topology.Molecule
            The second molecule to map
        """
        return self._molecule2

    def _ipython_display_(self):
        """
        It returns a representation of the mapping.

        Returns
        -------
        mapping_representation : a IPython display object
            Displayable RDKit molecules with mapping information
        """
        from IPython.display import display
        from peleffy.utils.toolkits import RDKitToolkitWrapper

        rdkit_toolkit = RDKitToolkitWrapper()

        mcs_mol = rdkit_toolkit.get_mcs(self.molecule1, self.molecule2,
                                        self._include_hydrogens,
                                        self._TIMEOUT)

        image = rdkit_toolkit.draw_mapping(self.molecule1, self.molecule2,
                                           mcs_mol, self._include_hydrogens)

        return display(image)



"""
This module defines the basic topology elements of a molecule.
"""


from simtk import unit


class _TopologyElement(object):
    """
    A wrapper for any topological element.
    """

    _name = None
    _writable_attrs = []
    _lambda_changeable = []

    class TopologyIterator(object):
        """
        An iterator for topological elements that iterates over their
        attributes in an ordered way.

        It is useful when writing topological elements to file.
        """

        def __init__(self, top_el):
            """
            It initiates a TopologyIterator object.

            Parameters
            ----------
            top_el : a TopologyElement object
                The topology element to iterate on.
            """
            self._index = int(0)
            self._top_el = top_el

        def __next__(self):
            """
            It returns the next item for the iteration.

            Returns
            -------
            attr_name : str
                The name of the attribute
            attr_value : float
                The value of the attribute
            """
            if self._index == len(self._top_el._writable_attrs):
                raise StopIteration

            attr_name = self._top_el._writable_attrs[self._index]
            attr_value = getattr(self._top_el, attr_name)
            self._index += 1

            return attr_name, attr_value

    @property
    def name(self):
        """
        The name that this topological element has.

        Returns
        -------
        name : str
            The name of the topological element
        """
        return self._name

    @property
    def n_writable_attrs(self):
        """
        The number of writable attributes this topological element has.

        Returns
        -------
        n_writable_attrs : int
            The number of writable attributes
        """
        return len(self._writable_attrs)

    def apply_lambda(self, attributes_to_modify, lambda_value,
                     reverse=False, final_state=None):
        """
        Given a lambda value, it modifies a set of attributes of
        this topological element. A lambda equal to 0 will keep the
        original attributes of this topological element. A lambda
        equal to 1 will remove completely the effects of the
        attributes. This behavior will be the opposite when reverse
        is set to True.

        Parameters
        ----------
        attributes_to_modify : list[str]
            The list attribute names that will be modified
        lambda_value : float
            A value between 0 and 1 that defined the lambda to
            apply to the attributes
        reverse : bool
            When set to true the effects of lambda will be the
            opposite
        final_state : a peleffy.topology.topology._TopologyElement object
            The topology element that represents the final state when
            lambda equals 1.0. Default is None, which means that the
            final state is not defined and therefore the topological
            element will disappear or will start from scratch
        """
        from peleffy.utils import Logger

        logger = Logger()

        # Check final_state
        if final_state is not None:
            if type(final_state) != type(self):
                logger.error([f'Final state must belong to the ' +
                              f'same topological element as ' +
                              f'the element that wants to be ' +
                              f'modified with a lambda value. ' +
                              f'It will not be changed.'])

                return

        if not reverse:
            lambda_value = 1.0 - lambda_value

        for attribute in attributes_to_modify:
            if attribute not in self._lambda_changeable:
                logger.error([f'Attribute {attribute} is not an '
                              f'attribute that can be changed with '
                              f'a lambda. It will not be changed.'])
                continue

            value = getattr(self, '_' + attribute)

            if value is not None:
                value = value * lambda_value

            if final_state is not None:
                final_value = getattr(final_state, '_' + attribute)

                if final_value is not None:
                    value = value + (1.0 - lambda_value) * final_value

            setattr(self, '_' + attribute, value)

    def __iter__(self):
        """
        It returns an instance of the TopologyIterator.

        Returns
        -------
        iterator : a TopologyIterator
            The TopologyIterator object
        """
        return self.TopologyIterator(self)

    def __repr__(self):
        """
        It returns the representation string of this topological element.

        Returns
        -------
        repr_string : str
            The representation string
        """
        repr_string = '{}('.format(self._name)
        attrs = [attr for attr in self]
        for attr_name, value in attrs[:-1]:
            repr_string += '{}={}, '.format(attr_name, value)
        repr_string += '{}={})'.format(*attrs[-1])

        return repr_string

    def __str__(self):
        """
        It returns the readable representation string of this topological
        element.

        Returns
        -------
        str_string : str
            The readable representation string
        """
        return self.__repr__()


class Atom(_TopologyElement):
    """
    It represents the properties of an atom.
    """

    _name = 'Atom'
    _writable_attrs = ['index', 'PDB_name', 'OPLS_type']
    _lambda_changeable = ['sigma', 'epsilon', 'charge', 'born_radius',
                          'SASA_radius', 'nonpolar_gamma', 'nonpolar_alpha']

    def __init__(self, index=-1, core=None, OPLS_type=None, PDB_name=None,
                 unknown=None, x=None, y=None, z=None, sigma=None,
                 epsilon=None, charge=None, born_radius=None, SASA_radius=None,
                 nonpolar_gamma=None, nonpolar_alpha=None, parent=None):
        """
        It initializes an Atom object.

        Parameters
        ----------
        index : int
            The index of the atom
        core : bool
            Whether atom is in the core or in a branch
        OPLS_type : str
            The OPLS type of the atom
        PDB_name : str
            The PDB name of the atom
        unknown : int
            The unknown value of the atom
        x : float
            The x coordinate of the atom
        y : float
            The y coordinate of the atom
        z : float
            The z coordinate of the atom
        sigma : simtk.unit.Quantity
            The sigma parameter of the atom
        epsilon : simtk.unit.Quantity
            The epsilon parameter of the atom
        charge : simtk.unit.Quantity
            The partial charge parameter of the atom
        born_radius : simtk.unit.Quantity
            The SGB born parameter radius of the atom
        SASA_radius : simtk.unit.Quantity
            The SASA radius parameter of the atom
        nonpolar_gamma : simtk.unit.Quantity
            The nonpolar gamma parameter of the atom
        nonpolar_alpha : simtk.unit.Quantity
            The nonpolar alpha parameter of the atom
        parent : peleffy.topology.Atom
            The parent of the atom
        """
        self._index = index
        self._core = core
        self._OPLS_type = OPLS_type
        self._PDB_name = PDB_name
        self._unknown = unknown
        self._x = x
        self._y = y
        self._z = z
        self._sigma = sigma
        self._epsilon = epsilon
        self._charge = charge
        self._born_radius = born_radius  # Rad. Non Polar SGB
        self._SASA_radius = SASA_radius  # Rad. Non Polar Type
        self._nonpolar_gamma = nonpolar_gamma  # SGB Non Polar gamma
        self._nonpolar_alpha = nonpolar_alpha  # SGB Non Polar type
        self._parent = parent

    def set_index(self, index):
        """
        It sets the index of the atom.

        Parameters
        ----------
        index : int
            The index of this Atom object
        """
        self._index = index

    def set_as_core(self):
        """It sets the atom as core."""
        self._core = True

    def set_as_branch(self):
        """It sets the atom as branch."""
        self._core = False

    def set_PDB_name(self, pdb_name):
        """
        It sets the PDB atom name.

        Parameters
        ----------
        pdb_name : str
            The PDB atom name to be set
        """
        self._PDB_name = pdb_name

    def set_parent(self, parent):
        """
        It sets the parent of the atom.

        Parameters
        ----------
        parent : peleffy.topology.Atom
            The parent of the atom
        """
        self._parent = parent

    def set_coords(self, coords):
        """
        It sets the coordinates of the atom.

        Parameters
        ----------
        coords : list
            The coordinates array to set to this Atom object
        """
        assert len(coords) == 3, '3D array is expected'

        self._x, self._y, self._z = coords

    def set_OPLS_type(self, OPLS_type):
        """
        It sets the OPLS type of the atom.

        Parameters
        ----------
        OPLS_type : str
            The OPLS type to set to this Atom object
        """
        self._OPLS_type = OPLS_type

    def set_sigma(self, sigma):
        """
        It sets the sigma of the atom.

        Parameters
        ----------
        sigma : float
            The sigma to set to this Atom object
        """
        self._sigma = sigma

    def set_epsilon(self, epsilon):
        """
        It sets the epsilon of the atom.

        Parameters
        ----------
        epsilon : float
            The epsilon to set to this Atom object
        """
        self._epsilon = epsilon

    def set_charge(self, charge):
        """
        It sets the charge of the atom.

        Parameters
        ----------
        charge : float
            The charge to set to this Atom object
        """
        self._charge = charge

    def set_born_radius(self, born_radius):
        """
        It sets the Born radius of the atom.

        Parameters
        ----------
        born_radius : float
            The Born radius to set to this Atom object
        """
        self._born_radius = born_radius

    def set_SASA_radius(self, SASA_radius):
        """
        It sets the SASA radius of the atom.

        Parameters
        ----------
        SASA_radius : float
            The SASA radius to set to this Atom object
        """
        self._SASA_radius = SASA_radius

    def set_nonpolar_gamma(self, nonpolar_gamma):
        """
        It sets the nonpolar gamma of the atom.

        Parameters
        ----------
        nonpolar_gamma : float
            The nonpolar gamma to set to this Atom object
        """
        self._nonpolar_gamma = nonpolar_gamma

    def set_nonpolar_alpha(self, nonpolar_alpha):
        """
        It sets the nonpolar alpha of the atom.

        Parameters
        ----------
        nonpolar_alpha : float
            The nonpolar alpha to set to this Atom object
        """
        self._nonpolar_alpha = nonpolar_alpha

    @property
    def index(self):
        """
        Atom's index.

        Returns
        -------
        index : int
            The index of this Atom object
        """
        return self._index

    @property
    def core(self):
        """
        Atom's core position.

        Returns
        -------
        core : bool
            Whether this Atom object is in the core or in a branch
        """
        return self._core

    @property
    def OPLS_type(self):
        """
        Atom's OPLS type.

        .. todo ::

           * Consider removing any reference to OPLS, if possible
             Otherwise, use SMIRks to find the best match

        Returns
        -------
        OPLS_type : str
            The OLPS type of this Atom object
        """
        return self._OPLS_type

    @property
    def PDB_name(self):
        """
        Atom's PDB name.

        .. todo ::

           * Consider removing any reference to OPLS, if possible
             Otherwise, use SMIRks to find the best match

        Returns
        -------
        PDB_name : str
            The PDB name of this Atom object
        """
        return self._PDB_name

    @property
    def unknown(self):
        """
        Atom's unknown int.

        .. todo ::

           * Review the actual purpose of this attribute in PELE

        Returns
        -------
        unknown : int
            The unknown int of this Atom object
        """
        return self._unknown

    @property
    def x(self):
        """
        Atom's x coordinate.

        Returns
        -------
        x : float
            The x coordinate of this Atom object
        """
        return self._x

    @property
    def y(self):
        """
        Atom's y coordinate.

        Returns
        -------
        y : float
            The y coordinate of this Atom object
        """
        return self._y

    @property
    def z(self):
        """
        Atom's z coordinate.

        Returns
        -------
        z : float
            The z coordinate of this Atom object
        """
        return self._z

    @property
    def sigma(self):
        """
        Atom's sigma.

        Returns
        -------
        sigma : simtk.unit.Quantity
            The sigma parameter of this Atom object
        """
        return self._sigma

    @property
    def epsilon(self):
        """
        Atom's epsilon.

        Returns
        -------
        epsilon : simtk.unit.Quantity
            The epsilon parameter of this Atom object
        """
        return self._epsilon

    @property
    def charge(self):
        """
        Atom's charge.

        Returns
        -------
        charge : simtk.unit.Quantity
            The charge parameter of this Atom object
        """
        return self._charge

    @property
    def born_radius(self):
        """
        Atom's born radius.

        Returns
        -------
        born_radius : simtk.unit.Quantity
            The SGB Born radius parameter of this Atom object
        """
        return self._born_radius

    @property
    def SASA_radius(self):
        """
        Atom's SASA radius.

        Returns
        -------
        SASA_radius : simtk.unit.Quantity
            The SASA radius parameter of this Atom object
        """
        return self._SASA_radius

    @property
    def nonpolar_gamma(self):
        """
        Atom's nonpolar gamma.

        Returns
        -------
        nonpolar_gamma : simtk.unit.Quantity
            The nonpolar gamma parameter of this Atom object
        """
        return self._nonpolar_gamma

    @property
    def nonpolar_alpha(self):
        """
        Atom's nonpolar alpha.

        Returns
        -------
        nonpolar_alpha : simtk.unit.Quantity
            The nonpolar alpha parameter of this Atom object
        """
        return self._nonpolar_alpha

    @property
    def parent(self):
        """
        Atom's parent.

        Returns
        -------
        parent : a peleffy.topology.Atom or None
            The parent of the atom. If it is the absolute parent, it
            will return a None
        """
        return self._parent


class DummyAtom(Atom):
    """
    It represents a dummy atom.
    """

    def __init__(self, index=-1, PDB_name='DUMM', parent=None):
        """
        It initializes a DummyAtom object.

        Parameters
        ----------
        index : int
            The index of the atom
        PDB_name : str
            The PDB name of the atom
        parent : peleffy.topology.Atom
            The parent of the atom
        """
        if parent is None:
            parent = self
        super().__init__(index, False, None, PDB_name, None, None, None, None,
                         None, None, None, None, None, None, None, parent)


class Bond(_TopologyElement):
    """
    It represents a bond in the topology.
    """

    _name = 'Bond'
    _writable_attrs = ['atom1_idx', 'atom2_idx', 'spring_constant', 'eq_dist']
    _lambda_changeable = ['spring_constant', 'eq_dist']

    def __init__(self, index=-1, atom1_idx=None, atom2_idx=None,
                 spring_constant=None, eq_dist=None):
        """
        It initiates a Bond object.

        Parameters
        ----------
        index : int
            The index of this Bond object
        atom1_idx : int
            The index of the first atom involved in this Bond
        atom2_idx : int
            The index of the second atom involved in this Bond
        spring_constant : simtk.unit.Quantity
            The spring constant of this Bond
        eq_dist : simtk.unit.Quantity
            The equilibrium distance of this Bond
        """
        self._index = index
        self._atom1_idx = atom1_idx
        self._atom2_idx = atom2_idx
        self._spring_constant = spring_constant
        self._eq_dist = eq_dist

    def set_index(self, index):
        """
        It sets the index of the bond.

        Parameters
        ----------
        index : int
            The index of this Bond object
        """
        self._index = index

    def set_atom1_idx(self, index):
        """
        It sets atom1's index.

        Parameters
        ----------
        index : int
            The index of the first atom involved in this Bond
        """
        self._atom1_idx = index

    def set_atom2_idx(self, index):
        """
        It sets atom2's index.

        Parameters
        ----------
        index : int
            The index of the second atom involved in this Bond
        """
        self._atom2_idx = index

    @property
    def index(self):
        """
        Bond's index.

        Returns
        -------
        index : int
            The index of this Bond object
        """
        return self._index

    @property
    def atom1_idx(self):
        """
        Bond's atom1 index.

        Returns
        -------
        atom1_idx : int
            The index of the first atom involved in this Bond object
        """
        return self._atom1_idx

    @property
    def atom2_idx(self):
        """
        Bond's atom2 index.

        Returns
        -------
        atom2_idx : int
            The index of the second atom involved in this Bond object
        """
        return self._atom2_idx

    @property
    def spring_constant(self):
        """
        Bond's spring constant.

        Returns
        -------
        spring_constant : simtk.unit.Quantity
            The spring constant of this Bond object
        """
        return self._spring_constant

    @property
    def eq_dist(self):
        """
        Bond's equilibrium distance.

        Returns
        -------
        eq_dist : simtk.unit.Quantity
            The equilibrium distance of this Bond object
        """
        return self._eq_dist


class Angle(_TopologyElement):
    """
    It represents an angle in the topology.
    """

    _name = 'Angle'
    _writable_attrs = ['atom1_idx', 'atom2_idx', 'atom3_idx',
                       'spring_constant', 'eq_angle']
    _lambda_changeable = ['spring_constant', 'eq_angle']

    def __init__(self, index=-1, atom1_idx=None, atom2_idx=None,
                 atom3_idx=None, spring_constant=None, eq_angle=None):
        """
        It initiates an Angle object.

        Parameters
        ----------
        index : int
            The index of this Angle object
        atom1_idx : int
            The index of the first atom involved in this Angle
        atom2_idx : int
            The index of the second atom involved in this Angle
        atom3_idx : int
            The index of the third atom involved in this Angle
        spring_constant : simtk.unit.Quantity
            The spring constant of this Angle
        eq_angle : simtk.unit.Quantity
            The equilibrium angle of this Angle
        """
        self._index = index
        self._atom1_idx = atom1_idx
        self._atom2_idx = atom2_idx
        self._atom3_idx = atom3_idx
        self._spring_constant = spring_constant
        self._eq_angle = eq_angle

    def set_index(self, index):
        """
        It sets the index of the angle.

        Parameters
        ----------
        index : int
            The index of this Angle object
        """
        self._index = index

    def set_atom1_idx(self, index):
        """
        It sets atom1's index.

        Parameters
        ----------
        index : int
            The index of the first atom involved in this Angle
        """
        self._atom1_idx = index

    def set_atom2_idx(self, index):
        """
        It sets atom2's index.

        Parameters
        ----------
        index : int
            The index of the second atom involved in this Angle
        """
        self._atom2_idx = index

    def set_atom3_idx(self, index):
        """
        It sets atom3's index.

        Parameters
        ----------
        index : int
            The index of the third atom involved in this Angle
        """
        self._atom3_idx = index

    @property
    def index(self):
        """
        Angle's index.

        Returns
        -------
        index : int
            The index of this Angle object
        """
        return self._index

    @property
    def atom1_idx(self):
        """
        Angle's atom1 index.

        Returns
        -------
        atom1_idx : int
            The index of the first atom involved in this Angle object
        """
        return self._atom1_idx

    @property
    def atom2_idx(self):
        """
        Angle's atom2 index.

        Returns
        -------
        atom1_idx : int
            The index of the second atom involved in this Angle object
        """
        return self._atom2_idx

    @property
    def atom3_idx(self):
        """
        Angle's atom3 index.

        Returns
        -------
        atom1_idx : int
            The index of the third atom involved in this Angle object
        """
        return self._atom3_idx

    @property
    def spring_constant(self):
        """
        Angle's spring constant.

        Returns
        -------
        spring_constant : simtk.unit.Quantity
            The spring constant of this Angle object
        """
        return self._spring_constant

    @property
    def eq_angle(self):
        """
        Angle's equilibrium angle.

        Returns
        -------
        eq_angle : simtk.unit.Quantity
            The equilibrium angle of this Angle object
        """
        return self._eq_angle


class Dihedral(_TopologyElement):
    """
    It represents a dihedral in the topology.

    It can be a proper or an improper dihedral.
    """

    _name = 'Dihedral'

    def __init__(self, index=-1, atom1_idx=None, atom2_idx=None,
                 atom3_idx=None, atom4_idx=None, periodicity=None,
                 prefactor=None, constant=None, phase=None):
        """
        It initiates an Dihedral object.

        Parameters
        ----------
        index : int
            The index of this Dihedral object
        atom1_idx : int
            The index of the first atom involved in this Dihedral
        atom2_idx : int
            The index of the second atom involved in this Dihedral
        atom3_idx : int
            The index of the third atom involved in this Dihedral
        atom4_idx : int
            The index of the fourth atom involved in this Dihedral
        periodicity : int
            The periodicity of this Dihedral
        prefactor : int
            The prefactor of this Dihedral
        constant : simtk.unit.Quantity
            The constant of this Dihedral
        phase : simtk.unit.Quantity
            The phase constant of this Dihedral
        """
        self._index = index
        self._atom1_idx = atom1_idx
        self._atom2_idx = atom2_idx
        self._atom3_idx = atom3_idx
        self._atom4_idx = atom4_idx
        self._periodicity = periodicity
        self._prefactor = prefactor
        self._constant = constant
        self._phase = phase

    def set_index(self, index):
        """
        It sets the index of the dihedral.

        Parameters
        ----------
        index : int
            The index of this Dihedral object
        """
        self._index = index

    def set_atom1_idx(self, index):
        """
        It sets atom1's index.

        Parameters
        ----------
        index : int
            The index of the first atom involved in this Dihedral
        """
        self._atom1_idx = index

    def set_atom2_idx(self, index):
        """
        It sets atom2's index.

        Parameters
        ----------
        index : int
            The index of the second atom involved in this Dihedral
        """
        self._atom2_idx = index

    def set_atom3_idx(self, index):
        """
        It sets atom3's index.

        Parameters
        ----------
        index : int
            The index of the third atom involved in this Dihedral
        """
        self._atom3_idx = index

    def set_atom4_idx(self, index):
        """
        It sets atom4's index.

        Parameters
        ----------
        index : int
            The index of the fourth atom involved in this Dihedral
        """
        self._atom4_idx = index

    def plot(self):
        """
        It plots this Dihedral as a function of phi angle.
        """
        from matplotlib import pyplot
        import numpy as np

        x = unit.Quantity(np.arange(0, np.pi, 0.1), unit=unit.radians)
        pyplot.plot(x, self.constant * (1 + self.prefactor
                                        * np.cos(self.periodicity * x
                                                 + self.phase)),
                    'r--')

        pyplot.show()

    @property
    def index(self):
        """
        Dihedral's index.

        Returns
        -------
        index : int
            The index of this Dihedral object
        """
        return self._index

    @property
    def atom1_idx(self):
        """
        Dihedral's atom1 index.

        Returns
        -------
        atom1_idx : int
            The index of the first atom involved in this Dihedral object
        """
        return self._atom1_idx

    @property
    def atom2_idx(self):
        """
        Dihedral's atom2 index.

        Returns
        -------
        atom1_idx : int
            The index of the second atom involved in this Dihedral object
        """
        return self._atom2_idx

    @property
    def atom3_idx(self):
        """
        Dihedral's atom3 index.

        Returns
        -------
        atom1_idx : int
            The index of the third atom involved in this Dihedral object
        """
        return self._atom3_idx

    @property
    def atom4_idx(self):
        """
        Dihedral's atom4 index.

        Returns
        -------
        atom1_idx : int
            The index of the fourth atom involved in this Dihedral object
        """
        return self._atom4_idx

    @property
    def periodicity(self):
        """
        Dihedral's periodicity.

        Returns
        -------
        periodicity : int
            The periodicity this Dihedral object
        """
        return self._periodicity

    @property
    def prefactor(self):
        """
        Dihedral's prefactor.

        Returns
        -------
        prefactor : int
            The prefactor this Dihedral object
        """
        return self._prefactor

    @property
    def constant(self):
        """
        Dihedral's constant.

        Returns
        -------
        constant : unit.simtk.Quantity
            The constant of this Dihedral object
        """
        return self._constant

    @property
    def phase(self):
        """
        Dihedral's phase constant.

        Returns
        -------
        phase : unit.simtk.Quantity
            The phase constant of this Dihedral object
        """
        return self._phase


class Proper(Dihedral):
    """
    It represents a proper dihedral in the topology.
    """

    _name = 'Proper'
    exclude = False
    _writable_attrs = ['atom1_idx', 'atom2_idx', 'atom3_idx', 'atom4_idx',
                       'constant', 'prefactor', 'periodicity', 'phase']
    _lambda_changeable = ['constant']

    def include_in_14_list(self):
        """
        It includes this proper dihedral in PELE's 1-4 list.
        """
        self.exclude = False

    def exclude_from_14_list(self):
        """
        It excludes this proper dihedral from PELE's 1-4 list by
        setting the index of the third atom to negative.
        """
        self.exclude = True


class Improper(Dihedral):
    """
    It represents an improper dihedral in the topology.
    """

    _name = 'Improper'
    _writable_attrs = ['atom1_idx', 'atom2_idx', 'atom3_idx', 'atom4_idx',
                       'constant', 'prefactor', 'periodicity']
    _lambda_changeable = ['constant']


class OFFDihedral(_TopologyElement):
    """
    It represents a dihedral in the Open Force Field's topology.
    """

    _name = 'OFFDihedral'
    _writable_attrs = ['atom1_idx', 'atom2_idx', 'atom3_idx', 'atom4_idx',
                       'periodicity', 'phase', 'k', 'idivf']
    _lambda_changeable = ['k']
    _to_PELE_class = Dihedral

    def __init__(self, index=-1, atom1_idx=None, atom2_idx=None,
                 atom3_idx=None, atom4_idx=None, periodicity=None,
                 phase=None, k=None, idivf=None):
        """
        It initiates an Dihedral object.

        Parameters
        ----------
        index : int
            The index of this Dihedral object
        atom1_idx : int
            The index of the first atom involved in this Dihedral
        atom2_idx : int
            The index of the second atom involved in this Dihedral
        atom3_idx : int
            The index of the third atom involved in this Dihedral
        atom4_idx : int
            The index of the fourth atom involved in this Dihedral
        periodicity : int
            The periodicity of this Dihedral
        phase : simtk.unit.Quantity
            The phase of this Dihedral
        k : simtk.unit.Quantity
            The constant of this Dihedral
        idivf : int
            The idivf of this Dihedral
        """
        self.index = index
        self.atom1_idx = atom1_idx
        self.atom2_idx = atom2_idx
        self.atom3_idx = atom3_idx
        self.atom4_idx = atom4_idx
        self.periodicity = periodicity
        self.phase = phase
        self.k = k
        self.idivf = idivf

    def _check_up(self):
        """
        It performs some parameter check ups.

        Raises
        ------
        AssertionError
            If any unexpected value is found
        """
        pass

    def to_PELE(self):
        """
        It converts this Open Force Field Dihedral object into a
        PELE-compatible one.

        .. todo ::

           * Fully cover all OpenFF dihedral parameters

        Returns
        -------
        PELE_dihedral : a Dihedral
            The PELE-compatible Dihedral object
        """
        if (self.periodicity is None or self.phase is None
                or self.k is None or self.idivf is None):
            return None

        try:
            self._check_up()
        except AssertionError as e:
            raise ValueError('Invalid value found: {}'.format(e))

        PELE_phase = self.phase

        if self.phase.value_in_unit(unit.degree) == 180:
            PELE_prefactor = -1
            PELE_phase = unit.Quantity(value=0.0, unit=unit.degree)
        else:
            PELE_prefactor = 1

        PELE_constant = self.k / self.idivf

        PELE_dihedral_kwargs = {'index': self.index,
                                'atom1_idx': self.atom1_idx,
                                'atom2_idx': self.atom2_idx,
                                'atom3_idx': self.atom3_idx,
                                'atom4_idx': self.atom4_idx,
                                'periodicity': self.periodicity,
                                'prefactor': PELE_prefactor,
                                'constant': PELE_constant,
                                'phase': PELE_phase}

        return self._to_PELE_class(**PELE_dihedral_kwargs)

    def plot(self):
        """
        It plots this Dihedral as a function of phi angle.
        """
        from matplotlib import pyplot
        import numpy as np

        x = unit.Quantity(np.arange(0, np.pi, 0.1), unit=unit.radians)
        pyplot.plot(x,
                    self.k * (1 + np.cos(self.periodicity * x - self.phase)),
                    'r--')

        pyplot.show()


class OFFProper(OFFDihedral):
    """
    It represents a proper dihedral in the Open Force Field's topology.
    """

    _name = 'OFFProper'
    _to_PELE_class = Proper

    def _check_up(self):
        """
        It performs some parameter check ups.

        .. todo ::

            * Periodicity can also equal 5 and currently it is not supported
             by PELE

        Raises
        ------
        AssertionError
            If any unexpected value is found
        """
        assert self.periodicity in (1, 2, 3, 4, 5, 6), 'Expected values ' \
            'for periodicity are 1, 2, 3, 4, 5 or 6, obtained ' \
            '{}'.format(self.periodicity)

        # proper's idivfs must always be 1 --> Apparently not: COC(O)OC, t84, t86
        # assert self.idivf == 1, 'The expected value for idivf is 1 ' \
        #     'for propers, obtained {}'.format(self.idivf)

        # The next version of PELE will be compatible with phase values
        # other than 0 and 180 degrees to fully cover all OpenFF dihedrals
        # assert self.phase.value_in_unit(unit.degree) in (0, 180), \
        #     'Expected values for phase are 0 or 180, obtained ' \
        #     '{}'.format(self.phase)


class OFFImproper(OFFDihedral):
    """
    It represents an improper dihedral in the Open Force Field's topology.
    """

    _name = 'OFFImproper'
    _to_PELE_class = Improper

    def _check_up(self):
        """
        It performs some parameter check ups.

        .. todo ::

            * Periodicity can also equal 5 and currently it is not supported
             by PELE

        Raises
        ------
        AssertionError
            If any unexpected value is found
        """
        assert self.periodicity in (1, 2, 3, 4, 5, 6), 'Expected values ' \
            'for periodicity are 1, 2, 3, 4, 5 or 6, obtained ' \
            '{}'.format(self.periodicity)

        assert self.phase.value_in_unit(unit.degree) in (0, 180), \
            'Expected values for phase are 0 or 180 in impropers, ' \
            'obtained {}'.format(self.phase)
